if LibStub then
  HazLoot = LibStub("AceAddon-3.0"):NewAddon("HazLoot", "AceConsole-3.0", "AceEvent-3.0", "AceComm-3.0")
else -- not in WoW
  HazLoot = {}
  self = HazLoot
  HazLoot.standalone = true
  function HazLoot:Print(message)
    print("[Core] "..message)
  end
  function HazLoot:Debug(message)
    print("[Core][DEBUG] "..message)
  end
  require("HazLootList")
  require("HSK")
  require("Bootstrap")
end

--TODO: Refactor the state machine to support strategies -- is it general enough?
--TODO: Separate out recipes, gems, pets etc. (maybe)
--TODO: Implement timers for bidding process
--TODO: Update loot lists when thresholds are changed.
--TODO: Present a frame for the suicide -> last call -> off-spec -> dust
--TODO: Impose sanity on the reset/fullreset/status/state commands

-- Constants
HazLoot.TRASH     = 0
HazLoot.COMMON    = 1
HazLoot.UNCOMMON  = 2
HazLoot.RARE      = 3
HazLoot.EPIC      = 4
HazLoot.LEGENDARY = 5
HazLoot.ARTIFACT  = 6

HazLoot.LootTypes = {"Common","Uncommon","Rare","Epic","Legendary","Artifact" }
HazLoot.LootTypes[HazLoot.TRASH]     = "Trash"
HazLoot.LootColours = { "ffffff", "1eff00", "0080ff", "b048f8", "ff8000", "ff0000",}
HazLoot.LootColours[HazLoot.TRASH]   = "9d9d9d"

HazLoot.STATE_IDLE           = 0
HazLoot.STATE_LOOTING        = 1
HazLoot.STATE_BID_OPEN       = 2
HazLoot.STATE_BID_LASTCALL   = 3
HazLoot.STATE_BID_CONFIRMED  = 4
HazLoot.STATE_ROLL_OPEN      = 5
HazLoot.STATE_ROLL_CONFIRMED = 6
HazLoot.STATE_SENDING_ITEM   = 7
HazLoot.STATES = {"Looting","Bid open","Bid - Last call","Bid confirming","Offspec roll open","Offspec roll confirmed","Sending item"}
HazLoot.STATES[0] = "Idle"

HazLoot.State = HazLoot.STATE_IDLE

HazLoot.PhatLewt     = {}   -- Table of current distributable loot
HazLoot.CurrentItem  = nil  -- loot distro sequence number of loot item being distributed
HazLoot.LastWinner   = {}   -- winner of last mediated distro
HazLoot.SuicideOrder = {}   -- record of all "hard" loot suicided for, by sequence suicided
HazLoot.Suicides     = {}   -- record of all "hard" loot suicided for by name (see format below)
HazLoot.Offspecs     = {}   -- record of all "hard" loot rolled for   ({name={link,link,link,},})
HazLoot.Duster       = nil  -- Designated disenchanter
HazLoot.PartyMembers = {}   -- Array of current members of the party/raid {name1,name2,...,nameN,}  (max 40)
HazLoot.PartyIndex   = {}   -- Inverted index into PartyMembers {[name1]=1,[name2]=2,...,[nameN]=n} (max 40)
HazLoot.LootList     = {}   -- Table of random loots by party member: {[rarity1]={[name1]=n1,[name2=n2],},[rarity2]={...},}
HazLoot.LeastLoot    = {}   -- Table of the least loot received by party members, by rarity: {[rarity1]=m1,[rarity2]=m2,..}
HazLoot.InRaid     = false
HazLoot.InParty    = true
HazLoot.Strategy   = HSK -- The bid strategy (SK currently, DKP, EPGP possible...)

local options = {
  name    = "HazLootConfig",
  handler = HazLoot,
  type    = 'group',
  args    = {
-- Visible in the Config UI
    debug = {
      type = "toggle",
      name = "Debug Messages",
      desc = "Enable debug messages in your chat.",
      get  = "IsDebug",
      set  = "ToggleDebug",
      width = "full",
      order = -1,
    },
    useSKList = { -- TODO: Replace with a Bid System selector
      type = "toggle",
      name = "SuicideKings",
      desc = "Use a Suicide Kings list for loot.",
      get  = "IsUsingSKList",
      set  = "ToggleUsingSKList",
      width = "full",
      order = 25,
    },
    usage = {
      type = "execute",
      name = "Help",
      desc = "Print help on the /hl commands to the chat window.",
      func = "Usage",
      width = "full",
      order = 30,
    },
    soft = {
      type = "select",
      name = "Soft Loot Threshold",
      desc = "Loot above this quality, but below the hard threshold will be randomly and evenly passed out to party/raid members.",
      values = HazLoot.LootTypes, 
      get    = "GetSoftThreshold",
      set    = "SetSoftThreshold",
      style  = "radio",
      order  = 11,
    },
    hard = {
      type = "select",
      name = "Hard Loot Threshold",
      desc = "Loot at or above this quality will be kept for mediated assignment.",
      values = HazLoot.LootTypes,
      get    = "GetHardThreshold",
      set    = "SetHardThreshold",
      style  = "radio",
      order  = 12,
    },
    status = { -- TODO: Combine report and status, and add a drop-down report target
      type = "execute",
      name = "Report in Chat",
      desc = "Report who's received what in your chat",
      func = "StatusReport",
      order = 20,
    },
    report = {
      type = "execute",
      name = "Report To Raid",
      desc = "Report who's received what to the raid/party",
      func = "RaidStatusReport",
      order = 21,
    },
-- HIDDEN (Command-line only)
    state = {
      type = "execute",
      name = "Current State",
      desc = "Report the current state.",
      func = "ReportCurrentState",
      hidden = true
    },
    reset = {
      type = "execute",
      name = "Reset Party List",
      desc = "Reset the party list, and clear random loot assignments",
      func = "ResetMemberList",
      hidden = true,
    },
    fullreset = {
      type = "execute",
      name = "Reset All Data",
      desc = "Reset the party list, clear random loot assignments, and clear the current loot list",
      func = "FullReset",
      hidden = true,
    },
    next = {
      type = "execute",
      name = "SuicideNext",
      desc = "Start suicide bidding for next item",
      func = "SuicideNext",
      hidden = true,
    },
    lastcall = {
      type = "execute",
      name = "SuicideLastCall",
      desc = "Last call for suicide bids on current item",
      func = "LastCall",
      hidden = true,
    },
    offspec = {
      type = "execute",
      name = "OffSpecRoll",
      desc = "Call for off-spec rolls for current item",
      func = "OffSpecRoll",
      hidden = true,
    },
    confirm = { -- This is a CYA announcement until the SK list code is written
      type  = "input",
      name  = "ConfirmWinner",
      desc  = "Confirm that you've read the SK list right",
      usage = "<Suicide Winner>",
      get   = "GetCurrentWinner",
      set   = "ConfirmWinner",
      hidden = true,
    },
    addalt = { -- This is a CYA announcement until the SK list code is written
      type  = "input",
      name  = "AddAlt",
      desc  = "Add an alt",
      usage = "<Main> <alt> <role>",
      get   = "GetAlt", -- ? useless
      set   = "AddAlt",
      hidden = true,
    },
    award = {
      type = "execute",
      name = "Award",
      desc = "Award loot to a winning player",
      func = "AwardCurrentItem",
      hidden = true,
    },
  },
}

-- Shortcuts for chat options
options.args.a    = options.args.award
options.args.c    = options.args.confirm
options.args.fr   = options.args.fullreset
options.args.lc   = options.args.lastcall
options.args.os   = options.args.offspec

local defaults = {
  profile = {
    debug         = true,
    useSKList     = true,
    soft          = HazLoot.UNCOMMON,
    hard          = HazLoot.EPIC,
  },
}

------- ACE AND ADDON EVENTS ---------

function HazLoot:OnInitialize()
  math.randomseed(os.time()) math.random() math.random() -- still not great, but better
  if not HazLoot.standalone then
    self.db = LibStub("AceDB-3.0"):New("HazLootDB", defaults)
    LibStub("AceConfig-3.0"):RegisterOptionsTable("HazLoot", options)
    self.optionsFrame = LibStub("AceConfigDialog-3.0"):AddToBlizOptions("HazLoot", "HazLoot")
    self:RegisterChatCommand("hazloot", "ChatCommand")
    self:RegisterChatCommand("hl", "ChatCommand")
  end
  if HazLoot:IsUsingSKList() then
    HazLoot:LoadSKList()
  end
  HazLoot:InitRarityTables()
end

function HazLoot:OnEnable()
  -- Called when the addon is enabled
  HazLoot:ResetMemberList(true) -- load member list, if there is one
  self:Print("I can haz loot!")

  self:RegisterEvent("LOOT_OPENED")
  self:RegisterEvent("LOOT_SLOT_CLEARED")
  self:RegisterEvent("PARTY_MEMBERS_CHANGED")
  self:RegisterEvent("RAID_ROSTER_UPDATE")
  self:RegisterEvent("UPDATE_INSTANCE_INFO")
end

function HazLoot:OnDisable()
  -- Called when the addon is disabled
end

--- WOW IN-GAME EVENTS ---

function HazLoot:UPDATE_INSTANCE_INFO()
  -- Check to see if this is a raid instance
  -- Check that you're NOT autolooting, if so, turn off autoloot
  -- Check for a duster - if no duster, warn every 30s (persist duster)
end

function HazLoot:PARTY_MEMBERS_CHANGED()
  self:GroupChange()
end

function HazLoot:RAID_ROSTER_UPDATE()
  self:GroupChange()
end

function HazLoot:LOOT_OPENED()
  HazLoot:LootThatBag()
end

function HazLoot:LOOT_SLOT_CLEARED(event, slot)
  HazLoot:LootCleared(slot)
end

-------- CORE ----------

function HazLoot:LootThatBag()
  local slots = GetNumLootItems()
  if 0 == slots then CloseLoot() return end -- Ignore empty bags
--[[ -- TODO: implement loot warning (with timer?)
  if 0 < (#HazLoot.PhatLewt) and HazLoot.State > HazLoot.STATE_LOOTING then
    return
  else
    if UnitName("target") and UnitLevel("target") < 0 then
      HazLoot:Announce("Boss Loot! "..UnitName("target").." opened for HazLooting.")
    elseif not UnitName("target") then -- probably container (this may be wrong)
      HazLoot:Announce("Boss Loot! HazLooting.")
    end
  end
]]--
  HazLoot.PhatLewt = nil-- Reset PhatLewt, because we're about to rescan
  HazLoot.PhatLewt = {} -- Reset PhatLewt, because we're about to rescan

  -- Iterate over the slots in the loot bag
  for i = 1,slots do
    if LootSlotIsCoin(i) then
      LootSlot(i)
    else
      local itemLink   = GetLootSlotLink(i)
      local id         = HazLoot:GetItemId(itemLink)
      local name, link, rarity, ilevel, mlevel, itype, subtype, stack, loc, texture = GetItemInfo(id)
      if (rarity < HazLoot:GetSoftThreshold()) or "Money" == itype then -- Just loot it
        LootSlot(i)
      elseif rarity < HazLoot:GetHardThreshold() then -- Random assign it (>= SoftThreshold)
        HazLoot:AssignRandomLoot(rarity,name,link,i,slots)
      else -- itemRarity >= HazLoot.HardThreshold, so throw it in PhatLewt array
        table.insert(HazLoot.PhatLewt, {itemId = id, itemName = name, itemLink = link, slot = i })
      end
    end
  end

  if 0 < (#HazLoot.PhatLewt) then
    -- re-sort, because may have re-opened bag
    table.sort(HazLoot.PhatLewt, function(a,b) return HazLootList:LootOrderComparator(a,b) end)
    if HazLoot.State == HazLoot.STATE_IDLE then
      HazLoot:Announce("Distributing loot in this order: ")
      for i,loot in ipairs(HazLoot.PhatLewt) do
        loot.roles = HazLootList:GetRolesOf(loot.itemName)
        HazLoot:Announce(i..". "..loot.itemLink.." ("..loot.roles..")")
      end
      HazLoot:SuicideNext()
    else
      for i,loot in ipairs(HazLoot.PhatLewt) do -- Still need to assign roles in table
        loot.roles = HazLootList:GetRolesOf(loot.itemName)
      end
    end
  end
end

function HazLoot:SuicideNext()
  if not HazLoot.PhatLewt or 0 == (#HazLoot.PhatLewt) then
    self:Debug("Everything's distributed - or should be. Current item is "..(HazLoot.CurrentItem or "nil"))
    return
  elseif HazLoot.State >= HazLoot.STATE_LOOTING then -- still in mid-bid
    local loot = HazLoot.PhatLewt[HazLoot.CurrentItem]
    if nil ~= loot then
      self:Debug("You haven't finished disposing of "..loot.itemLink)
    else
      self:Debug("I'm confused.")
    end
    return
  end
  for order,loot in pairs(HazLoot.PhatLewt) do
    HazLoot.CurrentItem = order
    HazLoot:Warn("PST with suicides for: "..order..". "..loot.itemLink.." ("..loot.roles..")")
    self:Print(" /hl |cff1eff00l|rast|cff1eff00c|rall to issue a last call for "..loot.itemLink)
    HazLoot.State = HazLoot.STATE_BID_OPEN
    break
  end

  -- start a 15s timer1
  -- listen to chat for ("bid","suicide","me","want","yes",...)
    -- (later: conditional bidding)
  -- if whisper string does not contain bid, whisper usage
  -- if bidder can not equip, whisper warning
  -- for each bid, bump timer1 out by max(5s,timer1)
  -- when timer1 is up, announce "Last Call (10s)"
  -- start a 10s timer
  -- for each new bid, bump timer2 out by max(5s,timer2) and announce
  -- when timer2 is up:
    -- intersect bidder list with SK List (for order)
    -- if all have looted: winner is winner
    -- if any have not yet looted: winner is highest such
    -- set hasLooted bit on winner
    -- announce winner (and number of loots this raid)
    -- distribute loot
    -- reorder SK List
end

function HazLoot:LastCall()
  if HazLoot.State ~= HazLoot.STATE_BID_OPEN then
    self:Print("No open bids for last call.")
    return
  end
  HazLoot.State = HazLoot.STATE_BID_LASTCALL
  HazLoot:Warn("Last Call for "..HazLoot.CurrentItem..". "..HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink)
  self:Print(" /hl |cff1eff00c|ronfirm  <name> to confirm the winner")
  self:Print(" /hl |cff1eff00o|rff|cff1eff00s|rpec to go to off-spec rolls")
end

function HazLoot:OffSpecRoll()
  if HazLoot.State ~= HazLoot.STATE_BID_LASTCALL then
    self:Print("Must issue last call for suicides first: /hl |cff1eff00l|rast|cff1eff00c|rall")
    return
  end
  HazLoot.State = HazLoot.STATE_ROLL_OPEN
  HazLoot:Warn("Roll offspec for "..HazLoot.CurrentItem..". "..HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink)
  self:Debug(" /hl |cff1eff00c|ronfirm  <name> to confirm the winner")
  -- Run an off-spec roll for an item:
    -- start a 15s timer1
    -- listen to rolls
    -- if roller must suicide, whisper warning... and go to suicide?
    -- if roller can not equip, whisper warning
    -- for each bid, bump timer1 out by 5s
    -- when timer1 is up, announce "Off-spec rolls for <item> end in 10s"
    -- start a 10s timer
    -- for each new bid, bump timer2 out by max(5s,timer2) and announce
    -- when timer2 is up:
      -- announce winner
      -- distribute loot
end

function HazLoot:Dust(item)
  self:Debug("Distributing <item> for dusting.")
end

function HazLoot:ConfirmWinner(info,winner)
  self:Debug("winner:"..winner)
  if "" == winner and HazLoot.LastWinner.name then
    if HazLoot.State == HazLoot.STATE_BID_LASTCALL then
      HazLoot.State = HazLoot.STATE_BID_CONFIRMED
    else --HazLoot.State ~= HazLoot.STATE_ROLL_OPEN
      HazLoot.State = HazLoot.STATE_ROLL_CONFIRMED
    end
    return
  end
  --[[
  if (HazLoot.InRaid and not UnitInRaid(winner)) or not UnitInParty(winner) then
    self:Debug(winner.." doesn't seem to be in the raid/party. Check your spelling.")
    return
  end
  ]]--
  if HazLoot.State ~= HazLoot.STATE_BID_LASTCALL and HazLoot.State ~= HazLoot.STATE_ROLL_OPEN then
    self:Print(" /hl |cff1eff00l|rast|cff1eff00c|rall to issue a last call before confirming or")
    self:Print(" /hl |cff1eff00o|rff|cff1eff00s|rpec  to go to off-spec rolls")
    return
  end
  local lootCandidate = nil
  for i=1,40 do -- Loop over all possible master loot candidates
    lootCandidate = GetMasterLootCandidate(i) -- Use GMLC() here, because consistency doesn't matter: loot's been assigned
    if lootCandidate and string.lower(winner) == string.lower(lootCandidate) then
      winner = lootCandidate
      HazLoot.LastWinner.index  = i
      HazLoot.LastWinner.name   = winner
      HazLoot.LastWinner.link   = HazLoot.PhatLewt[HazLoot.CurrentItem].itemLink
      HazLoot.LastWinner.slot   = HazLoot.PhatLewt[HazLoot.CurrentItem].slot
      if HazLoot.State == HazLoot.STATE_BID_LASTCALL then
        HazLoot:Announce("Suiciding "..HazLoot.LastWinner.link.." to "..winner.." - speak now if you think I've got the SK List wrong, or file a ticket later")
        HazLoot.State = HazLoot.STATE_BID_CONFIRMED
      else --HazLoot.State ~= HazLoot.STATE_ROLL_OPEN
        HazLoot:Announce("Awarding "..HazLoot.LastWinner.link.." to "..winner.." - speak now if you think I've misread the offspec rolls")
        HazLoot.State = HazLoot.STATE_ROLL_CONFIRMED
      end
      self:Print(" /hl |cffff1e00a|rward  <name> to confirm the winner")
      return
    end
  end
end

function HazLoot:AwardCurrentItem(info)
  if HazLoot.State ~= HazLoot.STATE_BID_CONFIRMED and HazLoot.State ~= HazLoot.STATE_ROLL_CONFIRMED then
    self:Print(" /hl |cff1eff00c|ronfirm  <name> to confirm the winner")
    return
  elseif HazLoot.State == HazLoot.STATE_BID_CONFIRMED then
    HazLoot.LastWinner.wintable = HazLoot.Suicides
    HazLoot.LastWinner.order = HazLoot.SuicideOrder
  else --HazLoot.State == HazLoot.STATE_ROLL_CONFIRMED then
    HazLoot.LastWinner.wintable = HazLoot.Offspecs
  end
  GiveMasterLoot(HazLoot.LastWinner.slot,HazLoot.LastWinner.index)
  HazLoot.State = HazLoot.STATE_SENDING_ITEM
end

function HazLoot:LootCleared(slot)
  if not HazLoot.CurrentItem then return end
  for itemId,loot in pairs(HazLoot.PhatLewt) do
    if slot == HazLoot.PhatLewt[HazLoot.CurrentItem].slot then
      local looter = HazLoot.LastWinner.name
      local link   = HazLoot.LastWinner.link
      HazLoot:Announce("Grats "..looter.."! "..link)
      -- HazLoot:GAnnounce("Grats "..looter.."! "..link)
      if not HazLoot.LastWinner.wintable[looter] then HazLoot.LastWinner.wintable[looter] = {} end
      table.insert(HazLoot.LastWinner.wintable[looter],link)
      if HazLoot.LastWinner.order then table.insert(HazLoot.LastWinner.order,link) HazLoot.LastWinner.order = nil end
      HazLoot.PhatLewt[HazLoot.CurrentItem] = nil
      HazLoot.CurrentItem = nil
      if 0 < (#HazLoot.PhatLewt) then
        HazLoot.State = HazLoot.STATE_LOOTING
      else
        HazLoot.State = HazLoot.STATE_IDLE
      end
      return
    end
  end
  self:Debug("Something may be wrong: I know a loot slot got cleared, but the item sent wasn't in the PhatLewt table.")
end

--- Assign loot to a random party member such that everyone gets N before anyone gets N+1 items of a give rarity.
-- @returns 
function HazLoot:AssignRandomLoot(itemRarity,itemName,itemLink,slot,slots)
  if HazLoot.standalone then self:Debug("[AssignRandomLoot] "..itemName.." in slot "..slot) end
  -- 1. Pick a random starting point in the list for items of this rarity
  -- 2. Cycle through through that list once until:
  -- 3a. EITHER find someone who has LeastLoot loots of that rarity, and award it
  -- 3b. OR find no nobody with LeastLoot lots of that rarity, award it to the random start and ++LeastLoot
  -- 4. If the attempted recipient is not a valid master loot candidate, skip them
  -- self:Debug(os.time())
  local tmpRand = math.random(40)
  local looterName
  local failures = {}
  for i = 1,#HazLoot.PartyMembers do
    local randomLooter = (i+tmpRand)%#HazLoot.PartyMembers + 1      -- +1 for LUA 1-based table indexes
    looterName = HazLoot.PartyMembers[randomLooter]
    if nil ~= looterName then
      if (HazLoot.LeastLoot[itemRarity] >= HazLoot.LootList[itemRarity][looterName]) then
        if HazLoot:GiveRandomLoot(itemRarity,itemLink,slot,looterName) then
          return {}
        else
          table.insert(failures,looterName)
        end
      end
    end
  end
  -- either everyone has at least LeastLoot items, or everyone who does not is not eligible
  if 0 == #failures then
    looterName = HazLoot.PartyMembers[tmpRand%#HazLoot.PartyMembers + 1]
    HazLoot.LeastLoot[itemRarity] = HazLoot.LeastLoot[itemRarity] +1
    if HazLoot:GiveRandomLoot(itemRarity,itemLink,slot,looterName) then
      self:Debug("Everyone should have received "..HazLoot.LeastLoot[itemRarity].." "..HazLoot.LootTypes[itemRarity].." lewt(s).")
      return {}
    end
  elseif #failures == #HazLoot.PartyMembers then
    self:Debug("Nobody was eligible for "..itemLink..", so I have no idea what to do. I'm probably buggered now.")
  else
    self:Debug("Assigning "..itemLink.." to someone with more than the minimum. Nobody else is eligible.")
    tmpRand = math.random(40) -- get a new seed
    for i = 1,#HazLoot.PartyMembers do
      local randomLooter = (i+tmpRand)%#HazLoot.PartyMembers + 1      -- +1 for LUA 1-based table indexes
      looterName = HazLoot.PartyMembers[randomLooter]
      if nil ~= looterName then
        if HazLoot:GiveRandomLoot(itemRarity,itemLink,slot,looterName) then
          return true
        end
      end
    end
  end
end

--- Give out loot randomly.
-- @param rarity The itemRarity of the loot
-- @param link   The itemLink of the loot
-- @param slot   The loot slot of the corpse being looted
-- @param looter The name of the party member being given the loot
-- @returns true if it was given out successfully, false otherwise
function HazLoot:GiveRandomLoot(rarity,link,slot,looter)
  for i=1,40 do -- Might as well loop over all possible slots
    if looter == GetMasterLootCandidate(i) then -- Use GMLC() here, because consistency doesn't matter: loot's been assigned
      GiveMasterLoot(slot,i)
      if nil == HazLoot.LootList[rarity][looter] then
        HazLoot.LootList[rarity][looter] = 1
      else
        HazLoot.LootList[rarity][looter] = HazLoot.LootList[rarity][looter] + 1
      end
      HazLoot:Announce(looter.." got "..link.." ("..HazLoot.LootList[rarity][looter].." "..HazLoot.LootTypes[rarity].."(s) this session).")
      return true
    end
  end
  -- Whisper this, instead of announcing?
  HazLoot:Announce("According to Blizzard, "..looter.." was not a valid master loot candidate for "..link)
  return false
end

--- Set up the loot tables 
function HazLoot:InitRarityTables()
  for rarity,name in ipairs(HazLoot.LootTypes) do
    if rarity >= HazLoot.GetSoftThreshold() then
      HazLoot.LootList[rarity]  = {}
      HazLoot.LeastLoot[rarity] = 0
    end
  end
end

--- Debug info
function HazLoot:ReportCurrentState(info)
  self:Print("State: "..HazLoot.STATES[HazLoot.State])
  self:Print("wintable is "..(HazLoot.LastWinner.wintable == HazLoot.Suicides and "Suicides" or "Offspecs"))
  self:Print("PhatLewt has "..(#HazLoot.PhatLewt).." elements")
end

function HazLoot:RaidStatusReport(info)
  HazLoot:GetStatusInfo(true)
end
function HazLoot:StatusReport(info)
  HazLoot:GetStatusInfo(false)
end

function HazLoot:GetStatusInfo(reportToRaid)
  local raider
  local tmpString = "In SK List: "..HSK:GetMasterListSize()..", present in raid: "..HSK:GetPresentCount()
  if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
  for i=1,40 do
    raider = HazLoot.PartyMembers[i]
    if (nil ~= raider) then
      tmpString = ""
      for rarity = HazLoot:GetSoftThreshold(),HazLoot:GetHardThreshold()-1 do
        tmpString = tmpString..HazLoot.LootList[rarity][raider].." "..HazLoot.LootTypes[rarity]..", "
        -- tmpString = tmpString..HazLoot:RarityMessage(HazLoot.LootList[rarity][raider].." "..HazLoot.LootTypes[rarity]..", ",rarity)
      end
      -- tmpString = tmpString..HazLoot:RarityMessage("to "..raider.." - ",0) -- trash colored if not present!
      tmpString = tmpString.."to "..raider.." - " -- trash colored if not present!
      if nil == HSK:IsPresent(raider) then
        tmpString = tmpString.." (unknown)"
      elseif HSK:IsPresent(raider) then
        tmpString = tmpString.." (present)"
      else
        tmpString = tmpString.." (not present)"
      end
      if HazLoot.Suicides[raider] then
        tmpString = tmpString.." Suicide(s): "
        for i,win in ipairs(HazLoot.Suicides[raider]) do
          tmpString = tmpString..i..". "..win
        end
      -- else tmpString = tmpString..HazLoot:RarityMessage(" (no suicides)",4)
      else tmpString = tmpString.." (no suicides)"
      end
      if HazLoot.Offspecs[raider] then
        tmpString = tmpString.." Offspec(s): "
        for i,win in ipairs(HazLoot.Offspecs[raider]) do
          tmpString = tmpString.." Offspec: "..win
          tmpString = tmpString..i..". "..win
        end
      -- else tmpString = tmpString..HazLoot:RarityMessage(" (no offspec)",4)
      else tmpString = tmpString.." (no offspec)"
      end
      if (reportToRaid) then HazLoot:Announce(tmpString) else HazLoot:Debug(tmpString) end
    end
  end
end

function HazLoot:LoadSKList() -- TODO: Make the bootstrap table an option
  if HazLoot:IsUsingSKList() then
    self:Debug("loading SK list")
    local mList = nil
    local aList = nil
    if self.db then 
      mList = self.db.profile.MasterList
      aList = self.db.profile.AltList
    end
    if mList then
      HSK:SetMasterList(mList, aList)
    else
      self:Debug("Bootstrapping Seraphim")
      HSK:Bootstrap("Seraphim")
    end
  else
    self:Debug("NOT loading SK List: it's disabled")
  end
end

function HazLoot:GetCurrentWinner()
end

function HazLoot:Usage()
  self:Print("Open a target with phat lewts to announce the distributable loots, then use /hazloot or /hl commands:")
  self:Print(" |cffff0000/hl next|r     - Open bidding on the next loot item.")
  self:Print(" |cffff0000/hl |rl|cffff0000ast|rc|cffff0000all|r - Last call for bids on the current loot item.")
  self:Print(" |cffff0000/hl |ro|cffff0000ff|rs|cffff0000pec|r  - Open item for off-spec rolls.")
  self:Print(" |cffff0000/hl |rc|cffff0000onfirm|r |cffff0000<name>|r       - Announce <name> as winner for confirmation.")
  self:Print(" |cffff0000/hl |ra|cffff0000ward|r             - Give current item to most recent confirmed winner.")
  self:Print("Configuration:")
  self:Print(" |cffff0000/hl|r - Bring up the horrible configuration UI.")
  self:Print(" |cffff0000/hl soft|r <rarity> - Set the soft threshold to <rarity>.")
  self:Print(" |cffff0000/hl hard|r <rarity> - Set the hard threshold to <rarity>.")
  self:Print("Help:")
  self:Print(" |cffff0000/hl help|r  - Display this message")
  self:Print(" |cffff0000/hl usage|r - Display this message")
end
------- PARTY LIST HANDLING ------

--TODO: look into Ace-Bucket
function HazLoot:GroupChange()
  if HazLoot.partyListLoaded then
    HazLoot:UpdateMemberList()
  else
    HazLoot:ResetMemberList(true) -- reload
  end
end

function HazLoot:UpdateMemberList()
  local looter = UnitName("player")
  local tmpCurrentParty = {}
  tmpCurrentParty[looter] = 0 -- Don't exclude yourself
  if (nil == HSK:MarkAsPresent(looter)) then self:Debug("Could not find "..looter.." in "..HazLoot.Strategy:GetName().." use /hl add or /hl addalt") end
  -- idempotent Add members still in the raid
  for i = 1,40 do
    looter = UnitName("raid"..i)
    if nil == looter then
      looter = UnitName("party"..i)
    end
    if (nil ~= looter) then
      tmpCurrentParty[looter] = i
      j = HazLoot:AddMemberToParty(looter)
      -- self:Debug("Found party member "..j..": "..looter)
    end
  end
  -- remove anybody who's no longer in the raid
  for i = 1,#HazLoot.PartyMembers do
    if nil == tmpCurrentParty[HazLoot.PartyMembers[i]] then
      HazLoot:UnloadPartyMember(i)
    end
  end
end

function HazLoot:UnloadPartyMember(index)
  -- TODO: make sure this is the correct behaviour with persistence
  if nil ~= HazLoot.PartyMembers[index] then
    HazLoot.PartyIndex[HazLoot.PartyMembers[index]] = nil
  end
  HazLoot.PartyMembers[index] = nil
  for rarity = HazLoot:GetHardThreshold(), HazLoot:GetSoftThreshold(), -1 do
    HazLoot.LootList[rarity][index] = nil
  end
end

function HazLoot:FullReset()
  HazLoot:ResetMemberList(true)
  HazLoot.PhatLewt = {} -- Table of current distributable loot
  HazLoot.CurrentItem  = nil
  HazLoot.LastWinner   = {}
  HazLoot.State = HazLoot.STATE_IDLE
end

function HazLoot:ResetMemberList(reload)
  for rarity = HazLoot:GetHardThreshold(), HazLoot:GetSoftThreshold(), -1 do
    HazLoot.LootList[rarity]  = nil
    HazLoot.LeastLoot[rarity] = 0
  end
  HazLoot.PartyMembers = nil
  HazLoot.PartyIndex   = nil
  HazLoot.PartyIndex   = {}
  HazLoot.PartyMembers = {}
  for i=HazLoot:GetSoftThreshold(), HazLoot:GetHardThreshold() do
    HazLoot.LootList[i] = nil
    HazLoot.LootList[i] = {}
  end
  HazLoot.partyListLoaded = false

  if reload then
    HazLoot.partyListLoaded = HazLoot:LoadPartyMembers()
  end
end

function HazLoot:LoadPartyMembers()
  if (0 >= GetNumRaidMembers()) then
    self:Debug("Not in raid, getting party members")
    HazLoot.InParty = true
  else
    self:Debug("In Raid, getting raid members")
    HazLoot.InRaid = true
  end
  local looter, j
  for i = 1,40 do
    looter = UnitName(i < 6 and "party"..i or "raid"..i)
    if (nil ~= looter) then
      j = HazLoot:AddMemberToParty(looter)
      -- self:Debug("Found party member "..j..": "..looter)
    end
  end
  -- Don't forget yourself!
  looter = UnitName("player")
  j = HazLoot:AddMemberToParty(looter)
  -- self:Debug("Found party member "..j..": "..looter)
  self:Debug(#HazLoot.PartyMembers..(HazLoot.InRaid and " raiders loaded " or " party members loaded"))
  return true
end

-- This function should be idempotent
function HazLoot:AddMemberToParty(looter)
  local i = HazLoot.PartyIndex[looter]
  if nil ~= i then return i end -- They're already in the list. Move on.
  -- otherwise, iterate over the max raid table size, looking for the first empty slot
  for i = 1,40 do -- 40 is largest possible raid size, so we can't need more of a table than that
    if nil == HazLoot.PartyMembers[i] then
      HazLoot.PartyMembers[i]    = looter
      HazLoot.PartyIndex[looter] = i
      for rarity,name in ipairs(HazLoot.LootTypes) do
        if (rarity >= HazLoot:GetSoftThreshold()) then
          HazLoot.LootList[rarity][looter] = HazLoot.LeastLoot[rarity]
        end
      end
      HSK:MarkAsPresent(looter)
      if (nil == HSK:MarkAsPresent(looter)) then
        self:Debug("Could not find "..looter.." in "..(HazLoot.Strategy and HazLoot.Strategy:GetName() or "unknown strategy,").." use /hl add or /hl addalt")
      else
        self:Debug(looter.." marked as present in "..(HazLoot.Strategy and HazLoot.Strategy:GetName() or "unknown strategy,"))
      end
      return i
    elseif looter == HazLoot.PartyMembers[i] then -- this path should now be redundant
      self:Debug("ERROR: "..looter.." is already party member #"..i.." but I should already know that")
      return i
    end
  end
  self:Debug("I couldn't find room for "..looter.." have you got more than 40 in the raid, or am I broken?")
  return -1
end

------- UTILITY FUNCTIONS ------

function HazLoot:RarityMessage(message,rarity)
  if (nil == HazLoot.LootColours[rarity]) then return message end
  return "|cff"..HazLoot.LootColours[rarity]..message.."|r"
end

--- Extract only the itemId from an itemLink using only Lua native functions
-- itemString: "item:itemId:enchantId:jewelId1:jewelId2:jewelId3:jewelId4:suffixId:uniqueId"
-- @params itemLink A WoW itemLink
-- @returns The numeric itemId of the itemLink
function HazLoot:GetItemId(itemLink)
  local found, _, itemString = string.find(itemLink, "^|c%x+|H(.+)|h%[.*%]")
  local itemId = string.sub(itemString,6,string.find(itemString,":",7)-1)
  return tonumber(itemId)
end

function HazLoot:Warn(message)
  if not HazLoot.standalone then
    SendChatMessage(message, "RAID_WARNING")
  else
    self:Debug("[RAID_WARNING]"..message)
  end
end

function HazLoot:GAnnounce(message)
  if not HazLoot.standalone then
    SendChatMessage(message, "GUILD")
  else
    self:Debug("[GUILD]"..message)
  end
end

function HazLoot:Announce(message)
  if HazLoot.InRaid then
    SendChatMessage(message, "RAID")
  elseif HazLoot.InParty and not HazLoot.standalone then
    SendChatMessage(message, "PARTY")
  else
    self:Debug(message)
  end
end

function HazLoot:Debug(message)
  if HazLoot:IsDebug() then
      self:Print(message)
  end
end

function HazLoot:ColouredRarity(rarity)
  return HazLoot:RarityMessage(HazLoot.LootTypes[rarity], rarity)
end
function HazLoot:PlainRarity(colouredRarity)
  -- "|cAARRGGBBRarity|r"
  -- "123456789012345678"
  return string.sub(colouredRarity,11,16)
end

------- GETTERS AND SETTERS FOR OPTIONS ---------
function HazLoot:AddAlt(info, params)
  --[[
  local sep  = string.find(params," ")
  local main = string.sub(params, 1, sep - 1)
  local params  = string.sub(params, sep + 1)
  sep = string.find(params," ")
  local alt  = string.sub(, 1, sep - 1)
  sep = string.find(params," ",sep + 1)
  local role  = string.sub(itemRoleString, 1, sep - 1)
  ]]--
  HazLoot:Debug("main:"..(main or "nil"))
end
function HazLoot:GetAlt()
  return nil
end

function HazLoot:IsDebug(info)
  return HazLoot.standalone and defaults.profile.debug or self.db.profile.debug
end
function HazLoot:ToggleDebug(info, newValue)
  self.db.profile.debug = newValue
end

function HazLoot:IsUsingSKList(info)
  return HazLoot.standalone and defaults.profile.useSKList or self.db.profile.useSKList
end
function HazLoot:ToggleUsingSKList(info, newValue)
  self.db.profile.useSKList = newValue
end

-- TODO: fix these so they show the coloured versions in the UI again, without booting the looter
function HazLoot:GetHardThreshold(info)
  return HazLoot.standalone and defaults.profile.hard or self.db.profile.hardThreshold
end
function HazLoot:SetHardThreshold(info,newValue)
  self.db.profile.hardThreshold = newValue
end
function HazLoot:GetSoftThreshold(info)
  if HazLoot.standalone then
    return tonumber(defaults.profile.soft)
  else
    return tonumber(self.db.profile.softThreshold)
  end
end
function HazLoot:SetSoftThreshold(info,newValue)
  self.db.profile.softThreshold = newValue
end

function HazLoot:GetLastWinner(info)
  if HazLoot.LastWinner.name then
    self:Debug("Last winner was "..HazLoot.LastWinner.name..", who won "..HazLoot.LastWinner.link)
    return HazLoot.LastWinner.name
  else
    self:Debug("No loot awarded yet")
  end
  return ""
end

function HazLoot:ChatCommand(input)
  if not input or input:trim() == "" then
    InterfaceOptionsFrame_OpenToCategory(self.optionsFrame)
  else
    LibStub("AceConfigCmd-3.0").HandleCommand(HazLoot, "hl", "HazLoot", input)
  end
end

-------- TESTING STUFF --------

if HazLoot.standalone then
  self:Print("Standalone")
  --TODO: write some tests here!
end

